/*
 * Copyright (c) 2008, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

using System.Text.RegularExpressions;
using System.IO;
using System.Collections;
using System.Runtime.InteropServices;

namespace MobiServer.Devices.Parallel.Control
{

    public partial class Base : MobiServer.Devices.ControlBase
    {
        #region Constants/Variables
        string serialNumber = "";
        private CheckBox[] checkBoxIOs;
        private ComboBox[] comboBoxIOs;

        int handle = 0;
        bool openFlag = false;
        public const string SETTING_HEADER = "setting_control"; //ft245rl
        bool commandSendFlag = true;

        enum GuiStates {Disabled, Close, Open };
        GuiStates guiState = GuiStates.Close;

        const string STR_OPEN_DEVICE = "foCXɐڑ܂:{0}";
        const string STR_OPEN_DEVICE_ERROR = "foCX̐ڑɎs܂:{0}";
        const string STR_CLOSE_DEVICE = "foCXؒf܂:{0}";
        const string STR_CLOSE_DEVICE_ERROR = "foCX̐ؒfɎs܂:{0}";
        const string STR_RESET_DEVICE = "foCXZbg܂:{0}";
        const string STR_RESET_DEVICE_ERROR = "foCX̃ZbgɎs܂:{0}";
        const string STR_SETTING_PORT = "IOݒɐ܂:{0:XX}";
        const string STR_SETTING_PORT_ERROR = "IOݒɎs܂:{0:XX}";
        const string STR_OUT_PORT = "o͂ɐ܂:{0:XX}";
        const string STR_OUT_PORT_ERROR = "o͂Ɏs܂:{0:XX}";
        const string STR_IN_PORT = "͂ɐ܂:{0:XX}";
        const string STR_IN_PORT_ERROR = "͂Ɏs܂:{0:XX}";

        #endregion

        #region Constructor/Dstructor

        public Base(string serial)
            : base(Data.Base.Header, GetLabel(serial))
        {
            InitializeComponent();
            InitComponentArray();
            this.serialNumber = serial;     
        }

        private void Base_Load(object sender, EventArgs e)
        {
            LoadSetting();
            ChangeGuiState(GuiStates.Close);
            SetIOMode();
        }

        void InitComponentArray()
        {
            checkBoxIOs = new System.Windows.Forms.CheckBox[8];
            checkBoxIOs[0] = checkBoxIO0;
            checkBoxIOs[1] = checkBoxIO1;
            checkBoxIOs[2] = checkBoxIO2;
            checkBoxIOs[3] = checkBoxIO3;
            checkBoxIOs[4] = checkBoxIO4;
            checkBoxIOs[5] = checkBoxIO5;
            checkBoxIOs[6] = checkBoxIO6;
            checkBoxIOs[7] = checkBoxIO7;

            comboBoxIOs = new System.Windows.Forms.ComboBox[8];
            comboBoxIOs[0] = comboBoxIO0;
            comboBoxIOs[1] = comboBoxIO1;
            comboBoxIOs[2] = comboBoxIO2;
            comboBoxIOs[3] = comboBoxIO3;
            comboBoxIOs[4] = comboBoxIO4;
            comboBoxIOs[5] = comboBoxIO5;
            comboBoxIOs[6] = comboBoxIO6;
            comboBoxIOs[7] = comboBoxIO7;
        }

        #endregion

        #region Utility Methods

        static string GetLabel(string serialNumber)
        {
            string str = String.Format("{0}:{1}", Data.Base.Header, serialNumber);
            return str;
        }

        public bool CheckAddress(string serialNumber)
        {
            if (Regex.Match(serialNumber, this.serialNumber, RegexOptions.IgnoreCase).Success)
                return true;
            else
                return false;
        }

        bool CheckAddress(string[] args)
        {
            string header = args[0];
            string serial = args[2];
           
            if (header == Data.Base.Header && (serial == "0" || CheckAddress(serial)))
                return true;
            else
                return false;
        }

        public bool GetEnabled()
        {
            return this.groupBox1.Enabled;
        }

        public void SetEnabled(bool enabled)
        {
            this.groupBox1.Enabled = enabled;
        }

        public string SerialNumber
        {
            get { return serialNumber; }
        }

        #endregion

        #region Attach/Dettach
        private void buttonDettach_Click(object sender, EventArgs e)
        {
            Dettach(this.groupBox1.Enabled);
            if (groupBox1.Enabled)
            {
                CreateServerWriteEvent(Data.States.Attach);
            }
            else
            {
                CreateServerWriteEvent(Data.States.Dettach);
            }
        }

        public void Dettach(bool flag)
        {
            if (flag)
            {
                buttonDettach.Text = "^";
                this.groupBox1.Enabled = false;
                this.SaveSetting();
            }
            else
            {
                this.groupBox1.Enabled = true;
                buttonDettach.Text = "x";
            }
        }

        #endregion

        
        #region Open/Close Serial Port

        private void buttonOpen_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void buttonReset_Click(object sender, EventArgs e)
        {
            Reset();
        }

        void ChangeGuiState(GuiStates state)
        {
            this.guiState = state;

            switch (state)
            {
                case GuiStates.Disabled:
                    buttonClose.Enabled = false;
                    buttonOpen.Enabled = false;
                    buttonReset.Enabled = false;
                    numericUpDownInterval.Enabled = false;
                    groupBoxPort.Enabled = false;
                    break;

                case GuiStates.Close:
                    buttonClose.Enabled = false;
                    buttonOpen.Enabled = true;
                    buttonReset.Enabled = false;
                    numericUpDownInterval.Enabled = true;
                    groupBoxPort.Enabled = true;
                    break;

                case GuiStates.Open:
                    buttonClose.Enabled = true;
                    buttonOpen.Enabled = false;
                    buttonReset.Enabled = true;
                    numericUpDownInterval.Enabled = false;
                    groupBoxPort.Enabled = true;
                    break;

            }
        }


        public unsafe bool Open()
        {
           bool result = false ;
           FT_STATUS status;
           int[] hDummy = new int[1]; 
           try
           {
              
               ChangeGuiState(GuiStates.Disabled);
               fixed (int* p = hDummy)
               {
                   status = FT_OpenEx(serialNumber, FT_OPEN_BY_SERIAL_NUMBER, p);
               }

               if (status == FT_STATUS.FT_OK)
               {
                   result = true;
                   handle = hDummy[0];
               }
           }
           catch (Exception ex)
           {
               PrintError(ex.ToString());
               buttonOpen.Enabled = true;
               result = false;
           }

           if (result)
           {
               PrintDebug(STR_OPEN_DEVICE, serialNumber);
               ChangeGuiState(GuiStates.Open);
               SetIOMode();
               SetData();
               timer1.Interval = (int)numericUpDownInterval.Value;
               timer1.Enabled = true;
              
           }
           else
           {
               PrintError(STR_OPEN_DEVICE_ERROR, serialNumber);
               ChangeGuiState(GuiStates.Close);
           }
           return result;
        }


        public bool Close()
        {
            bool result = false;
            FT_STATUS status;
            try
            {
                ChangeGuiState(GuiStates.Disabled);
                status = FT_Close(handle);
                if (status == FT_STATUS.FT_OK)
                {
                    result = true;
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
            }
            
            if (result)
            {
                PrintDebug(STR_CLOSE_DEVICE, serialNumber);
            }
            else
            {
                buttonOpen.Enabled = true;
                PrintError(STR_CLOSE_DEVICE_ERROR, serialNumber);
            }

            timer1.Enabled = false;
            ChangeGuiState(GuiStates.Close);
            return result;
        }

        public bool Reset()
        {
            bool result = false;
            FT_STATUS status;

            try
            {
                ChangeGuiState(GuiStates.Disabled);
                status = FT_Close(handle);
                if (status == FT_STATUS.FT_OK)
                {
                    result = true;
                }

            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
                //result = false;
            }
            

            if (result)
            {
                SetIOMode();
                PrintMessage(STR_RESET_DEVICE, serialNumber);
            }
            else
            {
                buttonOpen.Enabled = true;
                PrintError(STR_RESET_DEVICE_ERROR, serialNumber);
            }
            ChangeGuiState(GuiStates.Open);
            return result;
        }

        #endregion

        #region Save / Load
        string GetSettingFile()
        {
            string str = SETTING_DIR + @"\" + String.Format("{0}_{1}.txt", SETTING_HEADER, serialNumber);
            return str;

        }

        override public void SaveSetting()
        {
            StreamWriter fileStream = null;
            string file = GetSettingFile();

            try
            {
                fileStream = new StreamWriter(file, false);

                for (int i = 0; i < comboBoxIOs.Length; i++)
                {
                    fileStream.Write("{0}\t", comboBoxIOs[i].SelectedItem);
                }
                fileStream.WriteLine();


                for (int i = 0; i < checkBoxIOs.Length; i++)
                {
                    fileStream.Write("{0}\t", checkBoxIOs[i].Checked);
                }
                fileStream.WriteLine();
            }
            catch (IOException ex)
            {
                PrintError(ex.Message);
            }
            catch (Exception ex)
            {
                PrintError("Saving " + Path.GetFileName(file));
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        override public void LoadSetting()
        {
            StreamReader fileStream = null;
            string file = GetSettingFile();

            try
            {
                if (File.Exists(file))
                {

                    fileStream = new StreamReader(file);

                    string line = null;
                    string[] args = null;

                    line = fileStream.ReadLine();

                    args = line.Split("\t".ToCharArray());

                    for (int i = 0; i < comboBoxIOs.Length; i++)
                    {
                        comboBoxIOs[i].SelectedItem = args[i];
                    }

                    line = fileStream.ReadLine();
                    args = line.Split("\t".ToCharArray());
                    for (int i = 0; i < checkBoxIOs.Length; i++)
                    {
                        checkBoxIOs[i].Checked = bool.Parse(args[i]);
                    }
                }
            }
            catch (IOException ex)
            {
                PrintError(ex.Message);
            }
            catch (Exception ex)
            {
                PrintError("Loading " + Path.GetFileName(file));
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }


        #endregion

        #region Set I/O mode


        private void comboBoxIO0_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO1_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO2_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO3_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO4_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO5_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO6_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }

        private void comboBoxIO7_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetIOMode();
        }
   
        bool SetIOMode()
        {
            byte mask = (byte)0;
            FT_STATUS status;
            bool result = false;

            if (guiState != GuiStates.Open)
            {
                return false;
            }

            for (int i = 0; i < comboBoxIOs.Length; i++)
            {
                if (comboBoxIOs[i].SelectedItem.ToString() == Data.States.Out)
                {
                    mask |= (byte)(1 << i);
                    checkBoxIOs[i].Enabled = true;
                }
                else
                {
                    checkBoxIOs[i].Enabled = false;
                }
            }

            try
            {
                status = FT_SetBitMode(handle, mask, FT_SET_BITBANG_ASYNC);
                if (status == FT_STATUS.FT_OK)
                {
                    result = true;
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
            }

            if (result)
            {
                PrintDebug(STR_SETTING_PORT, mask);
            }
            else
            {
                PrintError(STR_SETTING_PORT_ERROR, mask);
            }

            return result;
        }

        #endregion

        #region Set/Get parallel data

        private void checkBoxIO0_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(0);
        }

        private void checkBoxIO1_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(1);

        }

        private void checkBoxIO2_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(2);
        }

        private void checkBoxIO3_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(3);
        }

        private void checkBoxIO4_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(4);
        }

        private void checkBoxIO5_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(5);
        }

        private void checkBoxIO6_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(6);
        }

        private void checkBoxIO7_CheckedChanged(object sender, EventArgs e)
        {
            OnCheckedChangedEvent(7);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            GetData();
        }

        private void numericUpDownInterval_ValueChanged(object sender, EventArgs e)
        {
            timer1.Interval = (int)numericUpDownInterval.Value;

        }

        void OnCheckedChangedEvent(int num)
        {
            if (comboBoxIOs[num].SelectedItem.ToString() == Data.States.Out)
            {
                SetData();
            }
            else
            {
                CreateServerWriteEvent(num, checkBoxIOs[num].Checked);
            }
        }

        unsafe bool SetData()
        {
            byte[] bytes = new byte[1];
            FT_STATUS status;
            uint writtenLength = 0;
            bool result = false;


            if (guiState != GuiStates.Open)
            {
                return false;
            }

            for (int i = 0; i < checkBoxIOs.Length; i++)
            {
                if (checkBoxIOs[i].Checked)
                {
                    bytes[0] |= (byte)(1 << i);
                }
            }

            try
            {
                fixed (byte* p = bytes)
                {
                    status = FT_Write(handle, p, (uint)bytes.Length, &writtenLength);
                }

                if (status == FT_STATUS.FT_OK)
                {
                    result = true;
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
            }

            if (result)
            {
                PrintDebug(STR_OUT_PORT, bytes[0]);
            }
            else
            {
                PrintError(STR_OUT_PORT_ERROR, bytes[0]);
            }

            return result;
        }

        unsafe bool GetData()
        {
            byte [] bytes= new byte[1];
            FT_STATUS status;
            bool result = false;

            if (guiState != GuiStates.Open)
            {
                return false;
            }

            try
            {
                fixed (byte* p = bytes)
                {
                    status = FT_GetBitMode(handle, p);
                }
                if (status == FT_STATUS.FT_OK)
                {
                    result = true;
                    for (int i = 0; i < checkBoxIOs.Length; i++)
                    {
                        if (comboBoxIOs[i].SelectedItem.ToString() == Data.States.In)
                        {
                            int flag = (int)((~bytes[0]) & 1 <<i);
                            
                            if ( flag >= 1)
                            {
                                checkBoxIOs[i].Checked = true;
                            }
                            else
                            {
                                checkBoxIOs[i].Checked = false;
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
            }

            if (result)
            {
                PrintDebug(STR_IN_PORT, bytes[0]);
            }
            else
            {
                PrintError(STR_IN_PORT_ERROR, bytes[0]);
                Close(); //error has occuered
            }

            return result;
        }

        #endregion

        #region Server In/Out
        void CreateServerWriteEvent(int port, bool flag)
        {
            int data = 0;
            if (flag)
                data = 1;

            Data.Base b = new Data.Base(Data.States.In,serialNumber,port,data);
            string[] args = b.ToArray();

            base.CreateServerWriteEvent(args);
        }

        override public void OnServerReadEvent(object sender, MobiServer.ServerEventArgs e)
        {
            try
            {
                if (CheckAddress(e.Args))
                {
                    Data.Base b = Data.Base.Parse(e.Args);

                    if (b != null)
                    {
                        if (b.State == Data.States.Dettach)
                        {
                            Dettach(true);
                        }
                        else if (b.State == Data.States.Out && b.Port >= 0 && b.Port < comboBoxIOs.Length)
                        //if (Regex.Match(b.State,  Data.States.Out, RegexOptions.IgnoreCase).Success)
                        {
                            if (comboBoxIOs[b.Port].SelectedItem.ToString() == Data.States.Out)
                            {
                                if (b.Data == 1)
                                    checkBoxIOs[b.Port].Checked = true;
                                else
                                    checkBoxIOs[b.Port].Checked = false;
                            }
                        }
                        else if (b.State == Data.States.Set && b.Port >= 0 && b.Port < comboBoxIOs.Length)
                        //else if (Regex.Match(b.State, Data.States.Set, RegexOptions.IgnoreCase).Success)
                        {
                            if (b.Data == 1)
                                comboBoxIOs[b.Port].SelectedItem = Data.States.Out;
                            else
                                comboBoxIOs[b.Port].SelectedItem = Data.States.In;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }

        void CreateServerWriteEvent(string state)
        {
            switch (state)
            {
                case Data.States.Attach:
                case Data.States.Dettach:
                    Data.Base b = new Data.Base(state, serialNumber);
                    string[] args = b.ToArray();
                    this.CreateServerWriteEvent(args);
                    break;
            }
        }

        #endregion

        #region Static methods
        public unsafe static ArrayList GetDeviceAll()
        {
            
            bool result = false;
            ArrayList deviceList = new ArrayList();

            int[] cnts = new int[1];
            FT_STATUS status;
            byte[] empty = new byte[1];

            try
            {
                fixed (int* p = cnts)
                {
                    fixed (byte* p2 = empty)
                    {
                        status = FT_ListDevices(p, p2, FT_LIST_BY_NUMBER_ONLY);
                    }
                }

                if (status != FT_STATUS.FT_OK || cnts[0] == 0)
                {
                    return null;
                }

                for (int i = 0; i < cnts[0]; i++)
                {
                    byte[] serialChars = new byte[20];
                    fixed (byte* p = serialChars)
                    {
                        status = FT_ListDevices(i, p, FT_LIST_BY_INDEX | FT_OPEN_BY_SERIAL_NUMBER);
                    }

                    if (status != FT_STATUS.FT_OK)
                    {
                        return null;
                    }
                    else
                    {
                        string s = System.Text.Encoding.ASCII.GetString(serialChars);
                        s = Regex.Replace(s, @"\0", "");
                        deviceList.Add(s);
                    }
                }
                return deviceList;
            }
            catch
            {
                return null;
            }
        }

        #endregion


        #region DLL import & constants for FT245RL

        enum FT_STATUS
        {
            FT_OK,
            FT_INVALID_HANDLE,
            FT_DEVICE_NOT_FOUND,
            FT_DEVICE_NOT_OPENED,
            FT_IO_ERROR,
            FT_INSUFFICIENT_RESOURCES,
            FT_INVALID_PARAMETER,
            FT_INVALID_BAUD_RATE,

            FT_DEVICE_NOT_OPENED_FOR_ERASE,
            FT_DEVICE_NOT_OPENED_FOR_WRITE,
            FT_FAILED_TO_WRITE_DEVICE,
            FT_EEPROM_READ_FAILED,
            FT_EEPROM_WRITE_FAILED,
            FT_EEPROM_ERASE_FAILED,
            FT_EEPROM_NOT_PRESENT,
            FT_EEPROM_NOT_PROGRAMMED,
            FT_INVALID_ARGS,
            FT_NOT_SUPPORTED,
            FT_OTHER_ERROR,
            FT_DEVICE_LIST_NOT_READY,
        };

        enum FT_DEVICE_TYPES
        {
            FT_DEVICE_BM,
            FT_DEVICE_AM,
            FT_DEVICE_100AX,
            FT_DEVICE_UNKNOWN,
            FT_DEVICE_2232C,
            FT_DEVICE_232R
        };


        //Flags for FT_ListDevices
        const uint FT_LIST_BY_NUMBER_ONLY = 0x80000000;
        const uint FT_LIST_BY_INDEX = 0x40000000;
        const uint FT_LIST_ALL = 0x20000000;

        //Flags for FT_OpenEx
        const int FT_OPEN_BY_SERIAL_NUMBER = 1;
        const int FT_OPEN_BY_DESCRIPTION = 2;
        const int FT_OPEN_BY_LOCATION = 4;

        //Flags for FT_SetBitMode
        const int FT_SET_RESET = 0x00;
        const int FT_SET_BITBANG_ASYNC = 0x01;
        const int FT_SET_MPSSE = 0x02;
        const int FT_SET_BITBANG_SYNC = 0x04;
        const int FT_SET_MCU_EMULATION = 0x08;
        const int FT_SET_FAST_SERIAL = 0x10;
        const int FT_SET_BITBANG_CBUS = 0x20;

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_Open(int iDevice, int* ftHandle);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern FT_STATUS FT_Close(int ftHandle);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern FT_STATUS FT_SetBaudRate(int ftHandle, uint dwBaudRate);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_ListDevices(int* pvArg1, byte* pvArg2, uint dwFlags);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_ListDevices(int pvArg1, byte* pvArg2, uint dwFlags);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_GetBitMode(int ftHandle, byte* pucMode);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern FT_STATUS FT_SetBitMode(int ftHandle, byte ucMask, byte ucMode);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_Read(int ftHandle, byte* lpBuffer, uint dwBytesToRead,
                                            uint* lpdwBytesReturned);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_Write(int ftHandle, byte* lpBuffer, uint dwBytesToWrite,
                                             uint* lpdwBytesWritten);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern FT_STATUS FT_ResetDevice(int ftHandle);

        [DllImport("ftd2xx", SetLastError = true)]
        static extern unsafe FT_STATUS FT_OpenEx(string pvArg1/* char* */, int dwFlags, int* ftHandle);

        #endregion

      

    }
}
