/*
 * Copyright (c) 2008, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace MobiServer.Devices.Parallel.Data
{
    public class States
    {
        public const string In = "In";
        public const string Out = "Out";
        public const string Set = "Set";
        public const string Attach = "Attach";
        public const string Dettach = "Dettach";
    }

    public class Base
    {
        public const string Header = "Parallel";  //USBParallel //FT245
        string state;
        string serialNumber;
        int port;
        int data;


        public Base(string state, string serial, int port, int data)
        {
            this.state = state;
            this.serialNumber = serial;
            this.port = port;
            this.data = data;
        }

        public Base(string state, string serial)
        {
            this.state = state;
            this.serialNumber = serial;
            this.port = 0;
            this.data = 0;
        }
        

        public string State
        {
            get { return this.state; }
            set { this.state = value; }
        }

        public string SerialNumber
        {
            get { return this.serialNumber; }
        }

       
        public int Port
        {
            get { return this.port; }
        }

        public int Data
        {
            get { return this.data; }
            set { this.data = value; }
        }

      
        public string[] ToArray()
        {
            string[] args = null;

            switch (State)
            {
                case States.Attach:
                case States.Dettach:
                    args = new string[3] { Header, State, SerialNumber };
                    break;
                case States.In:
                case States.Out:
                    args = new string[5] { Header, State, SerialNumber, Port.ToString(), Data.ToString()};
                    break;
            }
            return args;
        }


        static public Base Parse(string [] args)
        {
            try
            {
                Base b = null;
                string state = args[1];
                string serial = args[2];
                if (args.Length < 4)
                {
                    b = new Base(state, serial);
                }
                else
                {
                    int port = int.Parse(args[3]);
                    int data = int.Parse(args[4]);
                    b = new Base(state, serial, port, data);
                }

                return b;
            }
            catch (Exception ex)
            {
                return null;
            }
        }
    }
}

